#
# Copyright (C) 2020 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
API routes for Qemu nodes.
"""

import os

from fastapi import APIRouter, WebSocket, Depends, Body, Path, Response, status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import StreamingResponse
from uuid import UUID

from gns3server import schemas
from gns3server.compute.project_manager import ProjectManager
from gns3server.compute.qemu import Qemu
from gns3server.compute.qemu.qemu_vm import QemuVM

responses = {404: {"model": schemas.ErrorMessage, "description": "Could not find project or Qemu node"}}

router = APIRouter(responses=responses)


def dep_node(project_id: UUID, node_id: UUID) -> QemuVM:
    """
    Dependency to retrieve a node.
    """

    qemu_manager = Qemu.instance()
    node = qemu_manager.get_node(str(node_id), project_id=str(project_id))
    return node


@router.post(
    "",
    response_model=schemas.Qemu,
    status_code=status.HTTP_201_CREATED,
    responses={409: {"model": schemas.ErrorMessage, "description": "Could not create Qemu node"}},
)
async def create_qemu_node(project_id: UUID, node_data: schemas.QemuCreate) -> schemas.Qemu:
    """
    Create a new Qemu node.
    """

    qemu = Qemu.instance()
    node_data = jsonable_encoder(node_data, exclude_unset=True)
    vm = await qemu.create_node(
        node_data.pop("name"),
        str(project_id),
        node_data.pop("node_id", None),
        linked_clone=node_data.get("linked_clone", True),
        qemu_path=node_data.pop("qemu_path", None),
        console=node_data.pop("console", None),
        console_type=node_data.pop("console_type", "telnet"),
        aux=node_data.get("aux"),
        aux_type=node_data.pop("aux_type", "none"),
        platform=node_data.pop("platform", None),
    )

    for name, value in node_data.items():
        if hasattr(vm, name) and getattr(vm, name) != value:
            setattr(vm, name, value)

    return vm.asdict()


@router.get("/{node_id}", response_model=schemas.Qemu)
def get_qemu_node(node: QemuVM = Depends(dep_node)) -> schemas.Qemu:
    """
    Return a Qemu node.
    """

    return node.asdict()


@router.put("/{node_id}", response_model=schemas.Qemu)
async def update_qemu_node(node_data: schemas.QemuUpdate, node: QemuVM = Depends(dep_node)) -> schemas.Qemu:
    """
    Update a Qemu node.
    """

    node_data = jsonable_encoder(node_data, exclude_unset=True)
    # update the console first to avoid issue if updating console type
    node.console = node_data.pop("console", node.console)
    for name, value in node_data.items():
        if hasattr(node, name) and getattr(node, name) != value:
            await node.update_property(name, value)
    node.updated()
    return node.asdict()


@router.delete("/{node_id}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Delete a Qemu node.
    """

    await Qemu.instance().delete_node(node.id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/duplicate", response_model=schemas.Qemu, status_code=status.HTTP_201_CREATED)
async def duplicate_qemu_node(
        destination_node_id: UUID = Body(..., embed=True),
        node: QemuVM = Depends(dep_node)
) -> schemas.Qemu:
    """
    Duplicate a Qemu node.
    """

    new_node = await Qemu.instance().duplicate_node(node.id, str(destination_node_id))
    return new_node.asdict()


@router.post("/{node_id}/resize_disk", status_code=status.HTTP_204_NO_CONTENT)
async def resize_qemu_node_disk(node_data: schemas.QemuDiskResize, node: QemuVM = Depends(dep_node)) -> Response:

    await node.resize_disk(node_data.drive_name, node_data.extend)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/start", status_code=status.HTTP_204_NO_CONTENT)
async def start_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Start a Qemu node.
    """

    qemu_manager = Qemu.instance()
    hardware_accel = qemu_manager.config.settings.Qemu.enable_hardware_acceleration
    if hardware_accel and "-machine accel=tcg" not in node.options:
        pm = ProjectManager.instance()
        if pm.check_hardware_virtualization(node) is False:
            pass  # FIXME: check this
            # raise ComputeError("Cannot start VM with hardware acceleration (KVM/HAX) enabled because hardware virtualization (VT-x/AMD-V) is already used by another software like VMware or VirtualBox")
    await node.start()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/stop", status_code=status.HTTP_204_NO_CONTENT)
async def stop_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Stop a Qemu node.
    """

    await node.stop()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/reload", status_code=status.HTTP_204_NO_CONTENT)
async def reload_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Reload a Qemu node.
    """

    await node.reload()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/suspend", status_code=status.HTTP_204_NO_CONTENT)
async def suspend_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Suspend a Qemu node.
    """

    await node.suspend()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/resume", status_code=status.HTTP_204_NO_CONTENT)
async def resume_qemu_node(node: QemuVM = Depends(dep_node)) -> Response:
    """
    Resume a Qemu node.
    """

    await node.resume()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post(
    "/{node_id}/adapters/{adapter_number}/ports/{port_number}/nio",
    status_code=status.HTTP_201_CREATED,
    response_model=schemas.UDPNIO,
)
async def create_qemu_node_nio(
        *,
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        nio_data: schemas.UDPNIO,
        node: QemuVM = Depends(dep_node)
) -> schemas.UDPNIO:
    """
    Add a NIO (Network Input/Output) to the node.
    The port number on the Qemu node is always 0.
    """

    nio = Qemu.instance().create_nio(jsonable_encoder(nio_data, exclude_unset=True))
    await node.adapter_add_nio_binding(adapter_number, nio)
    return nio.asdict()


@router.put(
    "/{node_id}/adapters/{adapter_number}/ports/{port_number}/nio",
    status_code=status.HTTP_201_CREATED,
    response_model=schemas.UDPNIO,
)
async def update_qemu_node_nio(
        *,
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        nio_data: schemas.UDPNIO,
        node: QemuVM = Depends(dep_node)
) -> schemas.UDPNIO:
    """
    Update a NIO (Network Input/Output) on the node.
    The port number on the Qemu node is always 0.
    """

    nio = node.get_nio(adapter_number)
    if nio_data.filters:
        nio.filters = nio_data.filters
    if nio_data.suspend:
        nio.suspend = nio_data.suspend
    await node.adapter_update_nio_binding(adapter_number, nio)
    return nio.asdict()


@router.delete("/{node_id}/adapters/{adapter_number}/ports/{port_number}/nio", status_code=status.HTTP_204_NO_CONTENT)
async def delete_qemu_node_nio(
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        node: QemuVM = Depends(dep_node)
) -> Response:
    """
    Delete a NIO (Network Input/Output) from the node.
    The port number on the Qemu node is always 0.
    """

    await node.adapter_remove_nio_binding(adapter_number)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.post("/{node_id}/adapters/{adapter_number}/ports/{port_number}/capture/start")
async def start_qemu_node_capture(
        *,
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        node_capture_data: schemas.NodeCapture,
        node: QemuVM = Depends(dep_node)
) -> dict:
    """
    Start a packet capture on the node.
    The port number on the Qemu node is always 0.
    """

    pcap_file_path = os.path.join(node.project.capture_working_directory(), node_capture_data.capture_file_name)
    await node.start_capture(adapter_number, pcap_file_path)
    return {"pcap_file_path": str(pcap_file_path)}


@router.post(
    "/{node_id}/adapters/{adapter_number}/ports/{port_number}/capture/stop", status_code=status.HTTP_204_NO_CONTENT
)
async def stop_qemu_node_capture(
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        node: QemuVM = Depends(dep_node)
) -> Response:
    """
    Stop a packet capture on the node.
    The port number on the Qemu node is always 0.
    """

    await node.stop_capture(adapter_number)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get("/{node_id}/adapters/{adapter_number}/ports/{port_number}/capture/stream")
async def stream_pcap_file(
        adapter_number: int,
        port_number: int = Path(..., ge=0, le=0),
        node: QemuVM = Depends(dep_node)) -> StreamingResponse:
    """
    Stream the pcap capture file.
    The port number on the Qemu node is always 0.
    """

    nio = node.get_nio(adapter_number)
    stream = Qemu.instance().stream_pcap_file(nio, node.project.id)
    return StreamingResponse(stream, media_type="application/vnd.tcpdump.pcap")


@router.websocket("/{node_id}/console/ws")
async def console_ws(websocket: WebSocket, node: QemuVM = Depends(dep_node)) -> None:
    """
    Console WebSocket.
    """

    await node.start_websocket_console(websocket)


@router.post("/{node_id}/console/reset", status_code=status.HTTP_204_NO_CONTENT)
async def reset_console(node: QemuVM = Depends(dep_node)) -> Response:

    await node.reset_console()
    return Response(status_code=status.HTTP_204_NO_CONTENT)
