#
# Copyright (C) 2021 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
API routes for templates.
"""

import hashlib
import json

import logging

log = logging.getLogger(__name__)

from fastapi import APIRouter, Request, Response, HTTPException, Depends, Response, status
from typing import List
from uuid import UUID

from gns3server import schemas
from gns3server.db.repositories.templates import TemplatesRepository
from gns3server.services.templates import TemplatesService
from gns3server.db.repositories.rbac import RbacRepository

from .dependencies.authentication import get_current_active_user
from .dependencies.database import get_repository

responses = {404: {"model": schemas.ErrorMessage, "description": "Could not find template"}}

router = APIRouter(responses=responses)


@router.post("/templates", response_model=schemas.Template, status_code=status.HTTP_201_CREATED)
async def create_template(
    template_create: schemas.TemplateCreate,
    templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
    current_user: schemas.User = Depends(get_current_active_user),
    rbac_repo: RbacRepository = Depends(get_repository(RbacRepository))
) -> schemas.Template:
    """
    Create a new template.
    """

    template = await TemplatesService(templates_repo).create_template(template_create)
    template_id = template.get("template_id")
    await rbac_repo.add_permission_to_user_with_path(current_user.user_id, f"/templates/{template_id}/*")
    return template


@router.get("/templates/{template_id}", response_model=schemas.Template, response_model_exclude_unset=True)
async def get_template(
    template_id: UUID,
    request: Request,
    response: Response,
    templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
) -> schemas.Template:
    """
    Return a template.
    """

    request_etag = request.headers.get("If-None-Match", "")
    template = await TemplatesService(templates_repo).get_template(template_id)
    data = json.dumps(template)
    template_etag = '"' + hashlib.md5(data.encode()).hexdigest() + '"'
    if template_etag == request_etag:
        raise HTTPException(status_code=status.HTTP_304_NOT_MODIFIED)
    else:
        response.headers["ETag"] = template_etag
        return template


@router.put("/templates/{template_id}", response_model=schemas.Template, response_model_exclude_unset=True)
async def update_template(
    template_id: UUID,
    template_update: schemas.TemplateUpdate,
    templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
) -> schemas.Template:
    """
    Update a template.
    """

    return await TemplatesService(templates_repo).update_template(template_id, template_update)


@router.delete(
    "/templates/{template_id}",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_template(
        template_id: UUID,
        templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
        rbac_repo: RbacRepository = Depends(get_repository(RbacRepository))
) -> Response:
    """
    Delete a template.
    """

    await TemplatesService(templates_repo).delete_template(template_id)
    await rbac_repo.delete_all_permissions_with_path(f"/templates/{template_id}")
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get("/templates", response_model=List[schemas.Template], response_model_exclude_unset=True)
async def get_templates(
        templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
        current_user: schemas.User = Depends(get_current_active_user),
        rbac_repo: RbacRepository = Depends(get_repository(RbacRepository))
) -> List[schemas.Template]:
    """
    Return all templates.
    """

    templates = await TemplatesService(templates_repo).get_templates()
    if current_user.is_superadmin:
        return templates
    else:
        user_templates = []
        for template in templates:
            if template.get("builtin") is True:
                user_templates.append(template)
                continue
            template_id = template.get("template_id")
            authorized = await rbac_repo.check_user_is_authorized(
                current_user.user_id, "GET", f"/templates/{template_id}")
            if authorized:
                user_templates.append(template)
    return user_templates


@router.post("/templates/{template_id}/duplicate", response_model=schemas.Template, status_code=status.HTTP_201_CREATED)
async def duplicate_template(
        template_id: UUID, templates_repo: TemplatesRepository = Depends(get_repository(TemplatesRepository)),
        current_user: schemas.User = Depends(get_current_active_user),
        rbac_repo: RbacRepository = Depends(get_repository(RbacRepository))
) -> schemas.Template:
    """
    Duplicate a template.
    """

    template = await TemplatesService(templates_repo).duplicate_template(template_id)
    await rbac_repo.add_permission_to_user_with_path(current_user.user_id, f"/templates/{template_id}/*")
    return template
